#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <graphics.h>

#define incptr(p, n) ((void *)(((uintptr_t)(p)) + (n)))

#include "u_vga16.termfont.inc"

#define PXL(ctx, x, y) ((y) * ctx->pitch + (x))
#define STEP_X(ctx) 1
#define STEP_Y(ctx) (ctx)->pitch

void putpixel(
    gfx_context *ctx,
    uint64_t x, uint64_t y,
    uint32_t clr
) {
  if(x >= ctx->width || y >= ctx->height) return;
  uint32_t *fb = (uint32_t *)ctx->buffer;
  fb[PXL(ctx, x, y)] = clr;
}

void draw_line(
    gfx_context *ctx,
    uint64_t x0, uint64_t x1,
    uint64_t y0, uint64_t y1,
    uint32_t clr
) {
  int64_t dx = x1 > x0 ? x1 - x0 : x0 - x1;
  int64_t dy = y1 > y0 ? y1 - y0 : y0 - y1;
  int sx = x1 > x0 ? 1 : -1;
  int sy = y1 > y0 ? 1 : -1;
  uint64_t x = x0, y = y0;
  int64_t diff = dx - dy;
  uint32_t *fb = (uint32_t *)ctx->buffer;
  while(1) {
    fb[PXL(ctx, x, y)] = clr;
    if(x == x1 && y == y1) break;

    if((2*diff) > -dy) {
      diff -= dy;
      x += sx;
    } else {
      diff += dx;
      y += sy;
    }
  }
}

void draw_rect(
    gfx_context *ctx,
    uint64_t x, uint64_t y,
    uint64_t width, uint64_t height,
    uint32_t clr
) {
  uint32_t *fb = (uint32_t *)ctx->buffer;
  uint64_t l1 = PXL(ctx, x, y);
  uint64_t l2 = PXL(ctx, x, y + height);
  for(uint64_t _x = 0; _x <= width; _x++) {
    fb[l1] = fb[l2] = clr;
    l1 += STEP_X(ctx);
    l2 += STEP_X(ctx);
  }
  l1 = PXL(ctx, x, y);
  l2 = PXL(ctx, x + width, y);
  for(uint64_t _y = 0; _y <= height; _y++) {
    fb[l1] = fb[l2] = clr;
    l1 += STEP_Y(ctx);
    l2 += STEP_Y(ctx);
  }
}

void fill_rect(
    gfx_context *ctx,
    uint64_t x, uint64_t y,
    uint64_t width, uint64_t height,
    uint32_t clr
) {
  uint32_t *fb = (uint32_t *)ctx->buffer;
  uint64_t loc = PXL(ctx, x, y);
  for(uint64_t _y = 0; _y <= height; _y++) {
    for(uint64_t _x = 0; _x <= width; _x += STEP_X(ctx)) {
      fb[loc + _x] = clr;
    }
    loc += STEP_Y(ctx);
}
}

void putCharacter(
    gfx_context *ctx,
    uint64_t x, uint64_t y,
    uint32_t clr_fg, uint32_t clr_bg,
    char c
) {
  unsigned char *chr = c ? font[(int)c-0x20]: font[0];
  if(x >= ctx->width || y >= ctx->height) return;
  uint32_t *fb = (uint32_t *)ctx->buffer;
  uint64_t loc = y * ctx->pitch + x;
  for(int row = 0; row < 16; row++) {
    for(int col = 0; col < 8; col++) {
      fb[loc+col] = ((chr[row]>>(7-col))&0x1) ? clr_fg : clr_bg;
    }
    loc += ctx->pitch;
  }
}

void flip(gfx_context *ctx) {
  for(uint64_t y = 0; y < ctx->height; y++) {
    memcpy(
        incptr(ctx->addr, y*ctx->pitch*ctx->bpp),
        incptr(ctx->buffer, y*ctx->pitch*ctx->bpp),
        ctx->width*ctx->bpp
    );
  }
}

gfx_context *framebuffer_make_subcontext(
    gfx_context *ctx,
    uint64_t x, uint64_t y,
    uint64_t width, uint64_t height
) {
  gfx_context *out = malloc(sizeof(gfx_context));
  uint64_t loc = y * ctx->pitch + x;

  out->width = width;
  out->height = height;
  out->bpp = ctx->bpp;
  out->pitch = ctx->pitch;
  out->addr = incptr(ctx->addr, loc*ctx->bpp);
  out->size = ctx->pitch * height * ctx->bpp;
  out->buffer = calloc(1, out->size);

  return out;
}